# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from caterpillar.abc import _ContextLike

class StructException(Exception):
    context: _ContextLike
    def __init__(self, message: str, context: _ContextLike | None = None) -> None: ...

class DynamicSizeError(StructException): ...
class OptionError(StructException): ...
class ValidationError(StructException): ...
class UnsupportedOperation(StructException): ...
class InvalidValueError(StructException): ...
class StreamError(StructException): ...
class DelegationError(StructException): ...
class Stop(StructException): ...
