#ifndef CATERPILLAR_H
#define CATERPILLAR_H

// API includes
#include "caterpillar/caterpillarapi.h"

#include "caterpillar/arch.h"
#include "caterpillar/atom.h"
#include "caterpillar/context.h"
#include "caterpillar/lengthinfo.h"
#include "caterpillar/module.h"
#include "caterpillar/option.h"

#include "caterpillar/atoms/builtin/builtin.h"
#include "caterpillar/atoms/builtin/repeated.h"
#include "caterpillar/atoms/builtin/switch.h"
#include "caterpillar/atoms/builtin/conditional.h"
#include "caterpillar/atoms/builtin/offset.h"



#endif